from typing import Union

import numpy as np
import rasters as rt
from rasters import Raster

def calculate_SEBAL_soil_heat_flux(
        Rn: Union[Raster, np.ndarray], 
        ST_C: Union[Raster, np.ndarray], 
        NDVI: Union[Raster, np.ndarray], 
        albedo: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    This function calculates the soil heat flux (G) in the Surface Energy Balance Algorithm for Land (SEBAL) model.
    The formula used in the function is a simplification of the more complex relationship between these variables in the energy balance at the surface.
    
    Parameters:
    Rn (np.ndarray): Net radiation at the surface.
    ST_C (np.ndarray): Surface temperature in Celsius.
    NDVI (np.ndarray): Normalized Difference Vegetation Index, indicating the presence and condition of vegetation.
    albedo (np.ndarray): Measure of the diffuse reflection of solar radiation.
    
    Returns:
    np.ndarray: The soil heat flux (G), a key component in the energy balance.
    
    Reference:
    "Evapotranspiration Estimation Based on Remote Sensing and the SEBAL Model in the Bosten Lake Basin of China" [^1^][1]
    """
    # Empirical coefficients used in the calculation
    coeff1 = 0.0038
    coeff2 = 0.0074
    
    # Vegetation cover correction factor
    NDVI_correction = 1 - 0.98 * NDVI ** 4
    
    # Calculation of the soil heat flux (G)
    G = Rn * ST_C * (coeff1 + coeff2 * albedo) * NDVI_correction
    
    G = rt.clip(G, 0, None)

    return G



from typing import Union, Dict
import warnings
import numpy as np
import pandas as pd
from datetime import datetime

import rasters as rt
from rasters import Raster, RasterGeometry

from geos5fp import GEOS5FP

from .constants import *

from .meteorology_conversion import SVP_Pa_from_Ta_C

from .priestley_taylor import GAMMA_PA
from .priestley_taylor import delta_Pa_from_Ta_C

from .vegetation_conversion import LAI_from_NDVI
from .vegetation_conversion import SAVI_from_NDVI
from .vegetation_conversion import fAPAR_from_SAVI
from .vegetation_conversion import fIPAR_from_NDVI

from .partitioning import calculate_relative_surface_wetness
from .partitioning import calculate_green_canopy_fraction
from .partitioning import calculate_plant_moisture_constraint
from .partitioning import calculate_soil_moisture_constraint
from .partitioning import calculate_plant_temperature_constraint

from .partitioning import calculate_soil_net_radiation
from .partitioning import calculate_soil_latent_heat_flux
from .partitioning import calculate_canopy_latent_heat_flux
from .partitioning import calculate_interception

from .fAPARmax import load_fAPARmax
from .Topt import load_Topt

from .net_radiation.verma_net_radiation import process_verma_net_radiation
from .soil_heat_flux.calculate_SEBAL_soil_heat_flux import calculate_SEBAL_soil_heat_flux


def PTJPL(
        NDVI: Union[Raster, np.ndarray],
        ST_C: Union[Raster, np.ndarray] = None,
        emissivity: Union[Raster, np.ndarray] = None,
        albedo: Union[Raster, np.ndarray] = None,
        Rn: Union[Raster, np.ndarray] = None,
        Ta_C: Union[Raster, np.ndarray] = None,
        RH: Union[Raster, np.ndarray] = None,
        SWin: Union[Raster, np.ndarray] = None,
        G: Union[Raster, np.ndarray] = None,
        Topt: Union[Raster, np.ndarray] = None,
        fAPARmax: Union[Raster, np.ndarray] = None,
        geometry: Raster = None,
        datetime_UTC: datetime = None,
        GEOS5FP_connection: GEOS5FP = None,
        resampling: str = "nearest",
        delta_Pa: Union[Raster, np.ndarray, float] = None,
        gamma_Pa: Union[Raster, np.ndarray, float] = GAMMA_PA,
        epsilon=None,
        beta_Pa: float = BETA_PA,
        PT_alpha: float = PT_ALPHA,
        minimum_Topt: float = MINIMUM_TOPT,
        floor_Topt: bool = FLOOR_TOPT) -> Dict[str, np.ndarray]:
    results = {}

    if geometry is None and isinstance(NDVI, Raster):
        geometry = NDVI.geometry

    if Topt is None and geometry is not None:
        Topt = load_Topt(geometry)

    if fAPARmax is None and geometry is not None:
        fAPARmax = load_fAPARmax(geometry)

    if GEOS5FP_connection is None:
        GEOS5FP_connection = GEOS5FP()

    if Ta_C is None and geometry is not None and datetime_UTC is not None:
        Ta_C = GEOS5FP_connection.Ta_C(
            time_UTC=datetime_UTC,
            geometry=geometry,
            resampling=resampling
        )

    if Ta_C is None:
        raise ValueError("air temperature (Ta_C) not given")
    
    if RH is None and geometry is not None and datetime_UTC is not None:
        RH = GEOS5FP_connection.RH(
            time_UTC=datetime_UTC,
            geometry=geometry,
            resampling=resampling
        )

    if RH is None:
        raise ValueError("relative humidity (RH) not given")

    if Rn is None and albedo is not None and ST_C is not None and emissivity is not None:
        if SWin is None and geometry is not None and datetime_UTC is not None:
            SWin = GEOS5FP_connection.SWin(
                time_UTC=datetime_UTC,
                geometry=geometry,
                resampling=resampling
            )

        Rn_results = process_verma_net_radiation(
            SWin=SWin,
            albedo=albedo,
            ST_C=ST_C,
            emissivity=emissivity,
            Ta_C=Ta_C,
            RH=RH
        )

        Rn = Rn_results["Rn"]

    if Rn is None:
        raise ValueError("net radiation (Rn) not given")

    if G is None and Rn is not None and ST_C is not None and NDVI is not None and albedo is not None:
        G = calculate_SEBAL_soil_heat_flux(
            Rn=Rn,
            ST_C=ST_C,
            NDVI=NDVI,
            albedo=albedo
        )

    if G is None:
        raise ValueError("soil heat flux (G) not given")

    # calculate meteorology

    # calculate saturation vapor pressure in kPa from air temperature in celsius
    # floor saturation vapor pressure at 1
    SVP_Pa = SVP_Pa_from_Ta_C(Ta_C)

    # constrain relative humidity between 0 and 1
    RH = rt.clip(RH, 0, 1)

    # calculate water vapor pressure in Pascals from relative humidity and saturation vapor pressure
    Ea_Pa = RH * SVP_Pa

    # calculate vapor pressure deficit from water vapor pressure
    VPD_Pa = rt.clip(SVP_Pa - Ea_Pa, 0, None)

    # calculate relative surface wetness from relative humidity
    fwet = calculate_relative_surface_wetness(RH)

    # calculate vegetation values

    # convert normalized difference vegetation index to soil-adjusted vegetation index
    SAVI = SAVI_from_NDVI(NDVI)

    # calculate fraction of absorbed photosynthetically active radiation from soil-adjusted vegetation index
    fAPAR = fAPAR_from_SAVI(SAVI)

    # calculate fIPAR from NDVI
    fIPAR = fIPAR_from_NDVI(NDVI)

    # replace zero fIPAR with NaN
    fIPAR = np.where(fIPAR == 0, np.nan, fIPAR)

    # calculate green canopy fraction (fg) from fAPAR and fIPAR, constrained between zero and one
    fg = calculate_green_canopy_fraction(fAPAR, fIPAR)

    # calculate plant moisture constraint (fM) from fraction of photosynthetically active radiation, constrained between zero and one
    fM = calculate_plant_moisture_constraint(fAPAR, fAPARmax)

    # calculate soil moisture constraint from mean relative humidity and vapor pressure deficit, constrained between zero and one
    fSM = calculate_soil_moisture_constraint(RH, VPD_Pa, beta_Pa=beta_Pa)

    # apply corrections to optimum temperature

    if floor_Topt:
        # when Topt exceeds observed air temperature, then set Topt to match air temperature
        Topt = rt.where(Ta_C > Topt, Ta_C, Topt)

    Topt = rt.clip(Topt, minimum_Topt, None)

    # calculate plant temperature constraint (fT) from optimal phenology
    fT = calculate_plant_temperature_constraint(Ta_C, Topt)

    LAI = LAI_from_NDVI(NDVI)

    # calculate delta / (delta + gamma) term if it's not given
    if epsilon is None:
        # calculate delta if it's not given
        if delta_Pa is None:
            # calculate slope of saturation to vapor pressure curve in kiloPascal per degree Celsius
            delta_Pa = delta_Pa_from_Ta_C(Ta_C)

        # calculate delta / (delta + gamma)
        epsilon = delta_Pa / (delta_Pa + gamma_Pa)

    # soil evaporation

    # caluclate net radiation of the soil from leaf area index
    Rn_soil = calculate_soil_net_radiation(Rn, LAI)
    results["Rn_soil"] = Rn_soil

    # calculate soil evaporation (LEs) from relative surface wetness, soil moisture constraint,
    # priestley taylor coefficient, epsilon = delta / (delta + gamma), net radiation of the soil,
    # and soil heat flux
    LE_soil = calculate_soil_latent_heat_flux(Rn_soil, G, epsilon, fwet, fSM, PT_alpha)
    results["LE_soil"] = LE_soil

    # canopy transpiration

    # calculate net radiation of the canopy from net radiation of the soil
    Rn_canopy = Rn - Rn_soil
    results["Rn_canopy"] = Rn_canopy

    # calculate potential evapotranspiration (pET) from net radiation, and soil heat flux
    PET = PT_alpha * epsilon * (Rn - G)
    results["PET"] = PET

    # calculate canopy transpiration (LEc) from priestley taylor, relative surface wetness,
    # green canopy fraction, plant temperature constraint, plant moisture constraint,
    # epsilon = delta / (delta + gamma), and net radiation of the canopy
    LE_canopy = calculate_canopy_latent_heat_flux(
        Rn_canopy=Rn_canopy,
        epsilon=epsilon,
        fwet=fwet,
        fg=fg,
        fT=fT,
        fM=fM,
        PT_alpha=PT_alpha
    )

    results["LE_canopy"] = LE_canopy

    # interception evaporation

    # calculate interception evaporation (LEi) from relative surface wetness and net radiation of the canopy
    LE_interception = calculate_interception(
        Rn_canopy=Rn_canopy,
        epsilon=epsilon,
        fwet=fwet,
        PT_alpha=PT_alpha
    )

    results["LE_interception"] = LE_interception

    # combined evapotranspiration

    # combine soil evaporation (LEs), canopy transpiration (LEc), and interception evaporation (LEi)
    # into instantaneous evapotranspiration (LE)
    LE = LE_soil + LE_canopy + LE_interception

    # constrain instantaneous evapotranspiration between zero and potential evapotranspiration
    LE = np.clip(LE, 0, PET)
    results["LE"] = LE

    return results



from typing import Union
import numpy as np
import rasters as rt
from rasters import Raster

KPAR = 0.5
MIN_FIPAR = 0.0
MAX_FIPAR = 1.0
MIN_LAI = 0.0
MAX_LAI = 10.0

def FVC_from_NDVI(NDVI: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    Convert Normalized Difference Vegetation Index (NDVI) to Fractional Vegetation Cover (FVC).

    Parameters:
        NDVI (Union[Raster, np.ndarray]): Input NDVI data.

    Returns:
        Union[Raster, np.ndarray]: Converted FVC data.
    """
    NDVIv = 0.52  # +- 0.03
    NDVIs = 0.04  # +- 0.03
    FVC = rt.clip((NDVI - NDVIs) / (NDVIv - NDVIs), 0.0, 1.0)

    return FVC

def LAI_from_NDVI(
        NDVI: Union[Raster, np.ndarray],
        min_fIPAR: float = MIN_FIPAR,
        max_fIPAR: float = MAX_FIPAR,
        min_LAI: float = MIN_LAI,
        max_LAI: float = MAX_LAI) -> Union[Raster, np.ndarray]:
    """
    Convert Normalized Difference Vegetation Index (NDVI) to Leaf Area Index (LAI).

    Parameters:
        NDVI (Union[Raster, np.ndarray]): Input NDVI data.

    Returns:
        Union[Raster, np.ndarray]: Converted LAI data.
    """
    fIPAR = rt.clip(NDVI - 0.05, min_fIPAR, max_fIPAR)
    fIPAR = np.where(fIPAR == 0, np.nan, fIPAR)
    LAI = rt.clip(-np.log(1 - fIPAR) * (1 / KPAR), min_LAI, max_LAI)

    return LAI

def SAVI_from_NDVI(NDVI: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    Linearly calculates Soil-Adjusted Vegetation Index from ST_K.
    :param NDVI: normalized difference vegetation index clipped between 0 and 1
    :return: soil-adjusted vegetation index
    """
    return NDVI * 0.45 + 0.132

def fAPAR_from_SAVI(SAVI: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    Linearly calculates fraction of absorbed photosynthetically active radiation from soil-adjusted vegetation index.
    :param SAVI: soil adjusted vegetation index
    :return: fraction of absorbed photosynthetically active radiation
    """
    return rt.clip(SAVI * 1.3632 + -0.048, 0, 1)

def fIPAR_from_NDVI(NDVI: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    Calculate fraction of intercepted photosynthetically active radiation from normalized difference vegetation index
    :param NDVI: normalized difference vegetation index
    :return: fraction of intercepted photosynthetically active radiation
    """
    return rt.clip(rt.clip(NDVI, 0, 1) - 0.05, 0, 1)



from typing import Union
import numpy as np

from rasters import Raster

# psychrometric constant gamma in kiloPascal per degree Celsius
# same as value for ventilated (Asmann type) psychrometers, with an air movement of some 5 m/s
# http://www.fao.org/docrep/x0490e/x0490e07.htm
GAMMA_KPA = 0.0662  # kPa/C

# psychrometric constant gamma in Pascal per degree Celsius
GAMMA_PA = GAMMA_KPA * 1000

def delta_kPa_from_Ta_C(Ta_C: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    return 4098 * (0.6108 * np.exp(17.27 * Ta_C / (237.7 + Ta_C))) / (Ta_C + 237.3) ** 2

def delta_Pa_from_Ta_C(Ta_C: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    return delta_kPa_from_Ta_C(Ta_C) * 1000

def calculate_epsilon(delta: Union[Raster, np.ndarray], gamma: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    return delta / (delta + gamma)

def epsilon_from_Ta_C(Ta_C: Union[Raster, np.ndarray], gamma_Pa: Union[Raster, np.ndarray, float] = GAMMA_PA) -> Union[Raster, np.ndarray]:
    delta_Pa = delta_Pa_from_Ta_C(Ta_C)
    epsilon = calculate_epsilon(delta=delta_Pa, gamma=gamma_Pa)

    return epsilon



from typing import Union, Dict
import warnings
import numpy as np
import rasters as rt
from rasters import Raster

STEFAN_BOLTZMAN_CONSTANT = 5.67036713e-8  # SI units watts per square meter per kelvin to the fourth


def daylight_from_SHA(SHA_deg: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    This function calculates daylight hours from sunrise hour angle in degrees.
    """
    return (2.0 / 15.0) * SHA_deg


def sunrise_from_SHA(SHA_deg: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    This function calculates sunrise hour from sunrise hour angle in degrees.
    """
    return 12.0 - (SHA_deg / 15.0)


def solar_dec_deg_from_day_angle_rad(day_angle_rad: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    This function calculates solar declination in degrees from day angle in radians.
    """
    return (0.006918 - 0.399912 * np.cos(day_angle_rad) + 0.070257 * np.sin(day_angle_rad) - 0.006758 * np.cos(
        2 * day_angle_rad) + 0.000907 * np.sin(2 * day_angle_rad) - 0.002697 * np.cos(
        3 * day_angle_rad) + 0.00148 * np.sin(
        3 * day_angle_rad)) * (180 / np.pi)


def day_angle_rad_from_doy(doy: Union[Raster, np.ndarray]) -> Union[Raster, np.ndarray]:
    """
    This function calculates day angle in radians from day of year between 1 and 365.
    """
    return (2 * np.pi * (doy - 1)) / 365


def SHA_deg_from_doy_lat(doy: Union[Raster, np.ndarray, int], latitude: np.ndarray) -> Raster:
    """
    This function calculates sunrise hour angle in degrees from latitude in degrees and day of year between 1 and 365.
    """
    # calculate day angle in radians
    day_angle_rad = day_angle_rad_from_doy(doy)

    # calculate solar declination in degrees
    solar_dec_deg = solar_dec_deg_from_day_angle_rad(day_angle_rad)

    # convert latitude to radians
    latitude_rad = np.radians(latitude)

    # convert solar declination to radians
    solar_dec_rad = np.radians(solar_dec_deg)

    # calculate cosine of sunrise angle at latitude and solar declination
    # need to keep the cosine for polar correction
    sunrise_cos = -np.tan(latitude_rad) * np.tan(solar_dec_rad)

    # calculate sunrise angle in radians from cosine
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        sunrise_rad = np.arccos(sunrise_cos)

    # convert to degrees
    sunrise_deg = np.degrees(sunrise_rad)

    # apply polar correction
    sunrise_deg = rt.where(sunrise_cos >= 1, 0, sunrise_deg)
    sunrise_deg = rt.where(sunrise_cos <= -1, 180, sunrise_deg)

    return sunrise_deg

def process_verma_net_radiation(
        SWin: np.ndarray,
        albedo: np.ndarray,
        ST_C: np.ndarray,
        emissivity: np.ndarray,
        Ta_C: np.ndarray,
        RH: np.ndarray,
        cloud_mask: np.ndarray = None) -> Dict:
    results = {}

    # Convert surface temperature from Celsius to Kelvin
    ST_K = ST_C + 273.15

    # Convert air temperature from Celsius to Kelvin
    Ta_K = Ta_C + 273.15

    # Calculate water vapor pressure in Pascals using air temperature and relative humidity
    Ea_Pa = (RH * 0.6113 * (10 ** (7.5 * (Ta_K - 273.15) / (Ta_K - 35.85)))) * 1000
    
    # constrain albedo between 0 and 1
    albedo = np.clip(albedo, 0, 1)

    # calculate outgoing shortwave from incoming shortwave and albedo
    SWout = np.clip(SWin * albedo, 0, None)
    results["SWout"] = SWout

    # calculate instantaneous net radiation from components
    SWnet = np.clip(SWin - SWout, 0, None)

    # calculate atmospheric emissivity
    eta1 = 0.465 * Ea_Pa / Ta_K
    # atmospheric_emissivity = (1 - (1 + eta1) * np.exp(-(1.2 + 3 * eta1) ** 0.5))
    eta2 = -(1.2 + 3 * eta1) ** 0.5
    eta2 = eta2.astype(float)
    eta3 = np.exp(eta2)
    atmospheric_emissivity = np.where(eta2 != 0, (1 - (1 + eta1) * eta3), np.nan)

    if cloud_mask is None:
        # calculate incoming longwave for clear sky
        LWin = atmospheric_emissivity * STEFAN_BOLTZMAN_CONSTANT * Ta_K ** 4
    else:
        # calculate incoming longwave for clear sky and cloudy
        LWin = np.where(
            ~cloud_mask,
            atmospheric_emissivity * STEFAN_BOLTZMAN_CONSTANT * Ta_K ** 4,
            STEFAN_BOLTZMAN_CONSTANT * Ta_K ** 4
        )
    
    results["LWin"] = LWin

    # constrain emissivity between 0 and 1
    emissivity = np.clip(emissivity, 0, 1)

    # calculate outgoing longwave from land surface temperature and emissivity
    LWout = emissivity * STEFAN_BOLTZMAN_CONSTANT * ST_K ** 4
    results["LWout"] = LWout

    # LWnet = np.clip(LWin - LWout, 0, None)
    LWnet = np.clip(LWin - LWout, 0, None)

    # constrain negative values of instantaneous net radiation
    Rn = np.clip(SWnet + LWnet, 0, None)
    results["Rn"] = Rn

    return results

def daily_Rn_integration_verma(
        Rn: Union[Raster, np.ndarray],
        hour_of_day: Union[Raster, np.ndarray],
        doy: Union[Raster, np.ndarray] = None,
        lat: Union[Raster, np.ndarray] = None,
        sunrise_hour: Union[Raster, np.ndarray] = None,
        daylight_hours: Union[Raster, np.ndarray] = None) -> Raster:
    """
    calculate daily net radiation using solar parameters
    this is the average rate of energy transfer from sunrise to sunset
    in watts per square meter
    watts are joules per second
    to get the total amount of energy transferred, factor seconds out of joules
    the number of seconds for which this average is representative is (daylight_hours * 3600)
    documented in verma et al, bisht et al, and lagouARDe et al
    :param Rn:
    :param hour_of_day:
    :param sunrise_hour:
    :param daylight_hours:
    :return:
    """
    if daylight_hours is None or sunrise_hour is None and doy is not None and lat is not None:
        sha_deg = SHA_deg_from_doy_lat(doy, lat)
        daylight_hours = daylight_from_SHA(sha_deg)
        sunrise_hour = sunrise_from_SHA(sha_deg)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        Rn_daily = 1.6 * Rn / (np.pi * np.sin(np.pi * (hour_of_day - sunrise_hour) / (daylight_hours)))
    
    return Rn_daily



from os.path import join, abspath, dirname
import numpy as np
import rasters as rt
from rasters import Raster, RasterGeometry

def load_fAPARmax(geometry: RasterGeometry) -> Raster:
    SCALE_FACTOR = 0.0001
    filename = join(abspath(dirname(__file__)), "fAPARmax_mean_CMG_int16.tif")
    image = rt.clip(rt.Raster.open(filename, geometry=geometry, resampling="cubic") * SCALE_FACTOR, 0, None)
    image.nodata = np.nan

    return image
